// Copyright Contributors to the Pystring project.
// SPDX-License-Identifier: BSD-3-Clause
// https://github.com/imageworks/pystring/blob/master/LICENSE

#include "pystring.h"

// when not in header only mode include the implementations as non inline
// functions
#ifndef PYSTRING_HEADER_ONLY
#include "pystring_impl.h"
#endif
